from .exceptions import InvalidTransition

__all__ = ["FiniteStateMachineMixin", "BaseFiniteStateMachineMixin"]


class BaseFiniteStateMachineMixin:
    """Base Mixin to add a state_machine behavior.

    Represents the state machine for the object.

    The states and transitions should be specified in the following way:

    .. code-block:: python

        state_machine = {
           'some_state': '__all__'
           'another_state': ('some_state', 'one_more_state')
           'one_more_state': None
        }

    Requires the implementation of :code:`current_state` and :code:`set_state`
    """

    state_machine = None

    def current_state(self):
        """Returns the current state in the FSM."""
        raise NotImplementedError("Subclass must implement this method!")

    def set_state(self, state):
        """Update the internal state field.

        :param state: type depends on the definition of the states.
        :type state: str or int
        """
        raise NotImplementedError("Subclass must implement this method!")

    def can_change(self, next_state):
        """Validates if the next_state can be executed or not.

        It uses the state_machine attribute in the class.
        """
        valid_transitions = self.get_valid_transitions()

        if not valid_transitions:
            return False

        return next_state in valid_transitions

    def get_valid_transitions(self):
        """Return possible states to whom a product can transition.

        :returns: valid transitions
        :rtype: tuple or list
        """
        current = self.current_state()
        valid_transitions = self.state_machine[current]

        if valid_transitions == "__all__":
            return self.state_machine.keys()

        return self.state_machine[current]

    def on_before_change_state(self, previous_state, next_state, **kwargs):
        """Called before a state changes.

        :param previous_state: type depends on the definition of the states.
        :type previous_state: str or int
        :param next_state: type depends on the definition of the states.
        :type next_state: str or int
        """
        pass

    def on_change_state(self, previous_state, next_state, **kwargs):
        """Called after a state changes.

        :param previous_state: type depends on the definition of the states.
        :type previous_state: str or int
        :param next_state: type depends on the definition of the states.
        :type next_state: str or int
        """
        pass

    def change_state(self, next_state, **kwargs):
        """Performs a transition from current state to the given next state if
        possible.

        Callbacks will be exacuted before an after changing the state.
        Specific state callbacks will also be called if they are implemented
        in the subclass.

        :param next_state: where the state must go
        :type next_state: str or int
        :returns: new state.
        :rtype: str or int
        :raises InvalidTransition: If transitioning is not possible
        """
        previous_state = self.current_state()

        if not self.can_change(next_state):
            msg = "The transition from {0} to {1} is not valid".format(
                previous_state, next_state
            )
            raise InvalidTransition(msg)

        name = "pre_{0}".format(next_state)
        callback = getattr(self, name, None)
        if callback:
            callback(**kwargs)

        self.on_before_change_state(previous_state, next_state, **kwargs)

        self.set_state(next_state)

        name = "post_{0}".format(next_state)
        callback = getattr(self, name, None)
        if callback:
            callback(**kwargs)

        self.on_change_state(previous_state, next_state, **kwargs)
        return next_state


class FiniteStateMachineMixin(BaseFiniteStateMachineMixin):
    """A drop in implementation. Ready to be used.

    Replace :code:`FIELD_NAME` in order to automatically retrieve or set
    from a different field.

    In order to use with django, just add a field :code:`state`
    or as defined in :code:`FIELD_NAME`
    and remember to use :code:`change_state` instead of simply assigning it
    """

    FIELD_NAME = "state"

    def current_state(self):
        return getattr(self, self.FIELD_NAME)

    def set_state(self, state):
        setattr(self, self.FIELD_NAME, state)



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['fsm']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'fsmpy',
    'version': '2.0.0',
    'description': 'Minimal state machine',
    'long_description': '========\nOverview\n========\n\n.. start-badges\n\n.. list-table::\n    :stub-columns: 1\n\n    * - docs\n      - |docs|\n    * - tests\n      - | |travis|\n        | |codecov|\n    * - package\n      - |version| |wheel| |supported-versions| |supported-implementations|\n\n.. |docs| image:: https://readthedocs.org/projects/pyfsm/badge/?style=flat\n    :target: https://readthedocs.org/projects/pyfsm\n    :alt: Documentation Status\n\n.. |travis| image:: https://travis-ci.org/Woile/pyfsm.svg?branch=master\n    :alt: Travis-CI Build Status\n    :target: https://travis-ci.org/Woile/pyfsm\n\n.. |codecov| image:: https://codecov.io/github/Woile/pyfsm/coverage.svg?branch=master\n    :alt: Coverage Status\n    :target: https://codecov.io/github/Woile/pyfsm\n\n.. |version| image:: https://img.shields.io/pypi/v/fsmpy.svg?style=flat\n    :alt: PyPI Package latest release\n    :target: https://pypi.python.org/pypi/fsmpy\n\n.. |wheel| image:: https://img.shields.io/pypi/wheel/fsmpy.svg?style=flat\n    :alt: PyPI Wheel\n    :target: https://pypi.python.org/pypi/fsmpy\n\n.. |supported-versions| image:: https://img.shields.io/pypi/pyversions/fsmpy.svg?style=flat\n    :alt: Supported versions\n    :target: https://pypi.python.org/pypi/fsmpy\n\n.. |supported-implementations| image:: https://img.shields.io/pypi/implementation/fsmpy.svg?style=flat\n    :alt: Supported implementations\n    :target: https://pypi.python.org/pypi/fsmpy\n\n\n.. end-badges\n\nMinimal state machine\n\n* Free software: BSD license\n\n.. contents::\n    :depth: 2\n\nUsage\n=====\n\n.. code-block:: python\n\n    import fsm\n\n    class MyTasks(fsm.FiniteStateMachineMixin):\n        """An example to test the state machine.\n\n        Contains transitions to everywhere, nowhere and specific states.\n        """\n\n        state_machine = {\n            \'created\': \'__all__\',\n            \'pending\': (\'running\',),\n            \'running\': (\'success\', \'failed\'),\n            \'success\': None,\n            \'failed\': (\'retry\',),\n            \'retry\': (\'pending\', \'retry\'),\n        }\n\n        def __init__(self, state):\n            """Initialize setting a state."""\n            self.state = state\n\n        def on_before_pending(self):\n            print("I\'m going to a pending state")\n\n::\n\n    In [4]: m = MyTasks(state=\'created\')\n\n    In [5]: m.change_state(\'pending\')\n    I\'m going to a pending state\n    Out[5]: \'pending\'\n\n    In [6]: m.change_state(\'failed\')\n    ---------------------------------------------------------------------------\n    InvalidTransition                         Traceback (most recent call last)\n    <ipython-input-6-71d2461eee74> in <module>()\n    ----> 1 m.change_state(\'failed\')\n\n    ~/pyfsm/src/fsm/fsm.py in change_state(self, next_state, **kwargs)\n        90             msg = "The transition from {0} to {1} is not valid".format(previous_state,\n        91                                                                        next_state)\n    ---> 92             raise InvalidTransition(msg)\n        93\n        94         name = \'pre_{0}\'.format(next_state)\n\n    InvalidTransition: The transition from pending to failed is not valid\n\n\nThere are hooks that can be included before a state transition happens and after.\n\nfsm will look for these functions\n\n::\n\n    pre_<state_name>\n    post_<state_name>\n\nAnd will give them any extra argument given to :code:`change_state`\n\nE.g:\n\nRunning :code:`m.change_state(\'pending\', name=\'john\')` will trigger :code:`pre_pending(name=\'john\')`\n\n\nInstallation\n============\n\n::\n\n    pip install fsmpy\n\n\nDjango integration\n==================\n\n.. code-block:: python\n\n    import fsm\n    from django.db import models\n\n\n    class MyModel(models.Model, fsm.FiniteStateMachineMixin):\n        """An example to test the state machine.\n\n        Contains transitions to everywhere, nowhere and specific states.\n        """\n\n        CHOICES = (\n            (\'created\', \'CREATED\'),\n            (\'pending\', \'PENDING\'),\n            (\'running\', \'RUNNING\'),\n            (\'success\', \'SUCCESS\'),\n            (\'failed\', \'FAILED\'),\n            (\'retry\', \'RETRY\'),\n        )\n\n        state_machine = {\n            \'created\': \'__all__\',\n            \'pending\': (\'running\',),\n            \'running\': (\'success\', \'failed\'),\n            \'success\': None,\n            \'failed\': (\'retry\',),\n            \'retry\': (\'pending\', \'retry\'),\n        }\n\n        state = models.CharField(max_length=30, choices=CHOICES, default=\'created\')\n\n        def on_change_state(self, previous_state, next_state, **kwargs):\n            self.save()\n\n\nDocumentation\n=============\n\nhttps://pyfsm.readthedocs.org/\n\nDevelopment\n===========\n\nTo run the tests run::\n\n    tox\n\nNote, to combine the coverage data from all the tox environments run:\n\n.. list-table::\n    :widths: 10 90\n    :stub-columns: 1\n\n    - - Windows\n      - ::\n\n            set PYTEST_ADDOPTS=--cov-append\n            tox\n\n    - - Other\n      - ::\n\n            PYTEST_ADDOPTS=--cov-append tox\n',
    'author': 'Santiago Fraire Willemoes',
    'author_email': 'santiwilly@gmail.com',
    'url': 'https://github.com/Woile/pyfsm',
    'packages': packages,
    'package_data': package_data,
}


setup(**setup_kwargs)
